use std::{
    cell::RefCell,
    env,
    fmt::{self, Display},
    io::{stdout, IsTerminal, Write},
    marker::PhantomData,
};

use chrono::{DateTime, Local, TimeZone};
use owo_colors::{AnsiColors, OwoColorize};
use tracing::{
    field::{Field, Visit},
    Event, Level, Metadata, Subscriber,
};
use tracing_subscriber::{
    fmt::{format::Writer, MakeWriter},
    Layer,
};

pub struct GayFormattingLayer<
    S,
    W: for<'w> MakeWriter<'w> + 'static = fn() -> std::io::Stdout,
    Tz: TimeZone = Local,
> {
    timer: DateTime<Tz>,
    with_ansi: bool,
    colors: GayColors,
    make_writer: W,
    _inner: PhantomData<fn(S)>,
}

struct GayVisitor<'w> {
    writer: Writer<'w>,
    indent: usize,
    injected_message: Option<String>,
}

pub struct GayColors {
    pub crate_name: AnsiColors,
    pub trace: AnsiColors,
    pub debug: AnsiColors,
    pub info: AnsiColors,
    pub warn: AnsiColors,
    pub error: AnsiColors,
    pub timer: AnsiColors,
}

impl<S> Default for GayFormattingLayer<S> {
    fn default() -> Self {
        Self::new()
    }
}

impl<S, W> Layer<S> for GayFormattingLayer<S, W>
where
    S: Subscriber,
    W: for<'w> MakeWriter<'w> + 'static,
{
    fn on_event(&self, event: &Event<'_>, _ctx: tracing_subscriber::layer::Context<'_, S>) {
        thread_local! {
            static BUF: RefCell<String> = const { RefCell::new(String::new()) };
        }

        BUF.with(|buf: &RefCell<String>| {
            let borrow = buf.try_borrow_mut();
            let mut a;
            let mut b;
            let buf = match borrow {
                Ok(buf) => {
                    a = buf;
                    &mut *a
                }
                _ => {
                    b = String::new();
                    &mut b
                }
            };

            let res = self.format_event(event, tracing_subscriber::fmt::format::Writer::new(buf));

            if res.is_ok() {
                let mut writer = self.make_writer.make_writer();
                if let Err(fuck) = writer.write_all(buf.as_bytes()) {
                    eprintln!("[gay tracing formatter crate] unable to write event: {fuck}\n")
                }
            } else {
                eprintln!(
                    "[gay tracing formatter crate] unable to format event; name: {}; fields: {:?}",
                    event.metadata().name(),
                    event.fields()
                );
            }

            buf.clear();
        });
    }
}

impl<'w> Visit for GayVisitor<'w> {
    fn record_debug(&mut self, field: &Field, value: &dyn fmt::Debug) {
        match field.name() {
            "message" => {
                write!(self.writer, " {:?}", value).ok();
                if let Some(msg) = self.injected_message.take() {
                    write!(self.writer, "\n{}", msg).ok();
                }
            }
            _ => {
                if let Some(msg) = self.injected_message.take() {
                    write!(self.writer, "\n{}", msg).ok();
                }
                write!(
                    self.writer,
                    "\n{:>indent$}: {:?}",
                    field.name(),
                    value,
                    indent = self.indent - 1
                )
                .ok();
            }
        }
    }
}

impl Default for GayColors {
    fn default() -> Self {
        Self {
            crate_name: AnsiColors::Green,
            trace: AnsiColors::Magenta,
            debug: AnsiColors::Blue,
            info: AnsiColors::White,
            warn: AnsiColors::BrightYellow,
            error: AnsiColors::Red,
            timer: AnsiColors::BrightBlack,
        }
    }
}

impl<S> GayFormattingLayer<S> {
    pub fn new() -> Self {
        Self {
            timer: Local::now(),
            with_ansi: stdout().is_terminal(),
            colors: GayColors::default(),
            make_writer: std::io::stdout,
            _inner: PhantomData,
        }
    }
}

impl<S, W, Tz> GayFormattingLayer<S, W, Tz>
where
    W: for<'w> MakeWriter<'w> + 'static,
    Tz: TimeZone,
    Tz::Offset: Display,
{
    pub fn timer(self, timer: DateTime<Tz>) -> Self {
        let Self {
            with_ansi,
            colors,
            make_writer,
            _inner,
            ..
        } = self;
        Self {
            timer,
            with_ansi,
            colors,
            make_writer,
            _inner,
        }
    }

    pub fn with_ansi(self, with_ansi: bool) -> Self {
        let Self {
            timer,
            colors,
            make_writer,
            _inner,
            ..
        } = self;
        Self {
            timer,
            with_ansi,
            colors,
            make_writer,
            _inner,
        }
    }

    pub fn colors(self, colors: GayColors) -> Self {
        let Self {
            timer,
            with_ansi,
            colors: _,
            make_writer,
            _inner,
        } = self;
        Self {
            timer,
            with_ansi,
            colors,
            make_writer,
            _inner,
        }
    }

    pub fn make_writer(self, make_writer: W) -> Self {
        let Self {
            timer,
            with_ansi,
            colors,
            _inner,
            ..
        } = self;
        Self {
            timer,
            with_ansi,
            colors,
            make_writer,
            _inner,
        }
    }

    const TIME_LEN: usize = "YYYY-MM-DD HH-MM-SS".len();

    fn make_from_crate_string(&self, meta: &Metadata<'static>) -> Option<String> {
        static CURRENT_CRATE_NAME: &str = env!("CARGO_PKG_NAME");
        match meta.module_path() {
            Some(path) => path
                .split_once("::")
                .map(|(crate_name, _)| crate_name)
                .filter(|name| *name != CURRENT_CRATE_NAME)
                .map(|name| {
                    name.to_owned()
                        .colorize_if_true(self.with_ansi, self.colors.crate_name)
                })
                .map(|name| format!("{:>indent$} {name}", "from crate", indent = Self::TIME_LEN)),
            None => Some(format!(
                "{:>indent$}",
                "from unknown crate",
                indent = Self::TIME_LEN,
            )),
        }
    }

    fn format_event(
        &self,
        event: &Event,
        mut writer: tracing_subscriber::fmt::format::Writer,
    ) -> std::fmt::Result {
        let meta = event.metadata();
        write!(writer, "{} {}", self.datetime(), self.level(meta.level()))?;

        let mut v = GayVisitor::new(
            writer.by_ref(),
            Self::TIME_LEN,
            self.make_from_crate_string(meta),
        );

        event.record(&mut v);

        writeln!(writer)?;

        Ok(())
    }

    fn datetime(&self) -> String {
        self.timer
            .format("%F %T")
            .to_string()
            .colorize_if_true(self.with_ansi, self.colors.timer)
    }

    fn level(&self, level: &Level) -> String {
        let (icon, color) = match *level {
            Level::TRACE => ("TRACE", self.colors.trace),
            Level::DEBUG => ("DEBUG", self.colors.debug),
            Level::INFO => ("INFO ", self.colors.info),
            Level::WARN => ("WARN ", self.colors.warn),
            Level::ERROR => ("ERROR", self.colors.error),
        };
        icon.to_string().colorize_if_true(self.with_ansi, color)
    }
}

impl<'w> GayVisitor<'w> {
    pub fn new(writer: Writer<'w>, indent: usize, message: Option<String>) -> Self {
        GayVisitor {
            writer,
            indent,
            injected_message: message,
        }
    }
}

trait ColorizeIfTrue {
    fn colorize_if_true(self, cond: bool, color: AnsiColors) -> Self;
}

impl ColorizeIfTrue for String {
    fn colorize_if_true(self, cond: bool, color: AnsiColors) -> Self {
        if cond {
            self.color(color).to_string()
        } else {
            self
        }
    }
}
